package org.netkernel.lang.xrl.endpoint;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPathExpressionException;

import org.netkernel.layer0.meta.impl.IdentifierArgumentMetaImpl;
import org.netkernel.layer0.meta.impl.SourcedArgumentMetaImpl;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.util.RequestBuilder;
import org.netkernel.layer0.util.XMLReadable;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.layer0.util.XPath;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.util.Utils;
import org.w3c.dom.*;

/** New XRL which is driven from logical structure of the space in which it operates rather than
 * an explicit links definition
 * @author tab
 *
 */
public class XRL2Runtime extends StandardAccessorImpl
{
	// we look for control tags in either the default namespace or this namespace
	private static final String XRL_NS = "http://netkernel.org/xrl";
	private static final String ARGUMENT_ATT="argument-";
	
	public XRL2Runtime()
	{	declareThreadSafe();
		declareArgument(new SourcedArgumentMetaImpl("template",null,null,new Class[] {Document.class}));
		declareArgument(new IdentifierArgumentMetaImpl("mimetype",null,"application/xml" ));
		declareArgument(new SourcedArgumentMetaImpl("tolerant",null,null,new Class[] {Void.class}));
		declareSourceRepresentation(Document.class);
	}
	
	public void onSource(INKFRequestContext aHelper) throws Exception
	{
		INKFRequestReadOnly thisReq=aHelper.getThisRequest();
		
		String templateId = thisReq.getArgumentValue("template");
		INKFRequest templateReq = aHelper.createRequest(templateId);
		
		templateReq.setRepresentationClass(Node.class);
		Node templateRep = (Node)aHelper.issueRequest(templateReq);
		Document template = getMutableClone(templateRep);
		
		boolean tolerant=aHelper.getThisRequest().argumentExists("tolerant");
		
		recurse(template.getDocumentElement(),aHelper, tolerant);
		
		INKFResponse response=aHelper.createResponseFrom(template);
		String mimetype = thisReq.getArgumentValue("mimetype");
		if (mimetype!=null)
		{	response.setMimeType(mimetype);
		}
		else
		{	response.setMimeType("application/xml");
		}
	}
	
	private void recurse(Element aElement, INKFRequestContext aHelper, boolean tolerant) throws Exception
	{
		String xrlns = aElement.getAttribute("xmlns:xrl");
		if (xrlns.length()>0)
		{	aElement.removeAttribute("xmlns:xrl");
		}
		
		Element e=XMLUtils.getFirstChildElement(aElement);
		while (e!=null)
		{	Element next=XMLUtils.getNextSiblingElement(e);
			String ns=e.getNamespaceURI();
			if(ns!=null && ns.equals(XRL_NS))
			{	String name=e.getLocalName();
				if (name.equals("include"))
				{	processInclude(e,aHelper, tolerant);
				}
		        else if (name.equals("include-children"))
		        { //TODO rsk do you really want this - what is the usecase?
		        	processInclude(e,aHelper, tolerant);
		        }
				else if (name.equals("resolve"))
				{	processResolve(e,aHelper, tolerant);
				}
				else if (name.equals("eval"))
				{	processEval(e,aHelper, tolerant);
				}
			}
			else
			{	String inlineResolve=e.getAttribute("xrl:resolve");
				if (inlineResolve.length()>0)
				{	processInlineResolve(e,inlineResolve,aHelper, tolerant);
				}
				String inlineEval=e.getAttribute("xrl:eval");
				if (inlineEval.length()>0)
				{	processInlineEval(e,inlineEval,aHelper, tolerant);
				}
				recurse(e, aHelper, tolerant);
			}
			e=next;
		}		
	}

	private void processInclude(Element aElement, INKFRequestContext aHelper, boolean tolerant) throws Exception
	{	
		String target=null;
		try
		{	INKFRequest req;
			Element toReplace=aElement;
			boolean isInline=XMLUtils.getFirstChildElement(aElement)==null;			
			if (isInline)
			{	if(aElement.hasAttribute("identifier"))
				{	target=aElement.getAttribute("identifier");
					req =aHelper.createRequest(target);
				}
				else
				{	exceptionHandler(aHelper, tolerant, "EX_INCLUDE", "MSG_NO_BASE_OR_TARGET",aElement, null, target);
					return;
				}
				processArgumentAttributes(req, aElement);
			}
			else
			{	XMLReadable r = new XMLReadable(aElement);
				req=buildRequest(r, aHelper);
				
				String xpath=r.getTrimText("xrl:xpath");
				if (xpath.length()>0)
				{	List<Node> nodes=XPath.eval(xpath, aElement);
					if (nodes.size()==1 && nodes.get(0) instanceof Element)
					{	toReplace=(Element)nodes.get(0);
					}
					else
					{	exceptionHandler(aHelper, tolerant, "EX_INCLUDE", "MSG_BAD_XPATH",aElement, null, target);	
					}
				}
			}
			req.setRepresentationClass(Node.class);
			Node n = (Node)aHelper.issueRequest(req);
			if (n==null)
			{	exceptionHandler(aHelper, tolerant, "EX_INCLUDE", "MSG_NULL",aElement, null, target);
				return;
			}
			else if (n instanceof Document)
			{	n=((Document)n).getDocumentElement();
			}
			Element frag=(Element)aElement.getOwnerDocument().importNode(n, true);
			toReplace.getParentNode().replaceChild(frag, toReplace);
			if (toReplace!=aElement)
			{	aElement.getParentNode().removeChild(aElement);
			}
			recurse(frag, aHelper, tolerant);
		}
		catch (Exception e)
		{	exceptionHandler(aHelper, tolerant, "EX_INCLUDE", "MSG_EVAL",aElement, e, target);
		}
	}

	
	
	private void processResolve(Element aElement, INKFRequestContext aHelper, boolean tolerant) throws Exception
	{	
		String target=null;
		try
		{	INKFRequest req;
			XMLReadable r = new XMLReadable(aElement);
			
			req=buildRequest(r, aHelper);
			String identifier = req.getIdentifier();
			identifier=dropScheme(identifier);
			String xpath=r.getTrimText("xrl:xpath");
			processSetXPath(aElement,xpath,identifier);
			aElement.getParentNode().removeChild(aElement);
		}
		catch (Exception e)
		{	exceptionHandler(aHelper, tolerant, "EX_RESOLVE", "MSG_EVAL",aElement, e, target);
		}
	}

	private void processInlineResolve(Element aElement, String aResolve, INKFRequestContext aHelper, boolean tolerant) throws Exception
	{	
		String target="??";
		
		try
		{	target = getElementValue(aElement, aResolve,aHelper);
			if(target.indexOf("[[")>=0)
			{
				String oldtarget=null;
				int p1, p2=0;
				while((p1=target.indexOf("[["))>=0)
				{	p2=target.indexOf("]]");
					oldtarget=target;
					target=target.substring(p1+2,p2);
					INKFRequest req =aHelper.createRequest(target);
					processArgumentAttributes(req, aElement);
					String identifier=req.getIdentifier();
					identifier=dropScheme(identifier);
					target=oldtarget.substring(0, p1)+identifier+oldtarget.substring(p2+2);
				}
				aElement.removeAttribute("xrl:resolve");	
			}
			else
			{	aElement.removeAttribute("xrl:resolve");
				INKFRequest req =aHelper.createRequest(target);
				processArgumentAttributes(req, aElement);
				String identifier=req.getIdentifier();
				target=dropScheme(identifier);
			}
			setElementValue(aElement,aResolve, target);
		}
		catch (Exception e)
		{	exceptionHandler(aHelper, tolerant, "EX_RESOLVE", "MSG_EVAL",aElement, e, target);
		}
	}
	
	private void processEval(Element aElement, INKFRequestContext aHelper, boolean tolerant) throws Exception
	{	
		String target=null;
	
		try
		{	
			XMLReadable r = new XMLReadable(aElement);
			INKFRequest req;
			req=buildRequest(r,aHelper);
			req.setRepresentationClass(String.class);
			String evalResult = (String)aHelper.issueRequest(req);
			if (evalResult==null)
			{	exceptionHandler(aHelper, tolerant, "EX_EVAL", "MSG_NULL",aElement, null, target);
				return;
			}
			String xpath=r.getTrimText("xrl:xpath");
			processSetXPath(aElement,xpath,evalResult);
			aElement.getParentNode().removeChild(aElement);
		}
		catch (Exception e)
		{	exceptionHandler(aHelper, tolerant, "EX_EVAL", "MSG_EVAL",aElement, e, target);
		}
	}
	
	private void processInlineEval(Element aElement, String aEval, INKFRequestContext aHelper, boolean tolerant) throws Exception
	{	String target = "??";
		try
		{	target=getElementValue(aElement, aEval,aHelper);
			aElement.removeAttribute("xrl:eval");	
			INKFRequest req =aHelper.createRequest(target);
			processArgumentAttributes(req, aElement);
			req.setRepresentationClass(String.class);
			String evalResult = (String)aHelper.issueRequest(req);
			if (evalResult==null)
			{	exceptionHandler(aHelper, tolerant, "EX_EVAL", "MSG_NULL",aElement, null, target);
				return;
			}
			setElementValue(aElement, aEval, evalResult);
		}
		catch (Exception e)
		{	exceptionHandler(aHelper, tolerant, "EX_EVAL", "MSG_EVAL",aElement, e, target);
		}
	}
	
	private void exceptionHandler(INKFRequestContext aHelper, boolean tolerant, String aException, String aMessage, Node aElement, Exception e, String target ) throws Exception
	{	if(tolerant)
		{	aHelper.logFormatted(INKFLocale.LEVEL_WARNING, aMessage, XMLUtils.getPathFor(aElement), e, target );
			if (e!=null)
			{	aHelper.logRaw(INKFLocale.LEVEL_WARNING, Utils.throwableToString(e));
			}
		}
		else
		{	throw aHelper.createFormattedException(aException, aMessage, XMLUtils.getPathFor(aElement), e, target );
		}	
	}

	private String getElementValue(Element aElement, String aIdentifier, INKFRequestContext aContext) throws Exception
	{	String result;
		if (aIdentifier.equals("text"))
		{	result = XMLUtils.getText(aElement);
		}
		else
		{	Attr attr=aElement.getAttributeNode(aIdentifier);
			if (attr!=null)
			{	result = attr.getNodeValue();
			}
			else
			{	throw aContext.createFormattedException("EX_ATTRIBUTE_NOT_FOUND", "MSG_ATTRIBUTE_NOT_FOUND", null, null, aIdentifier);
			}
				
		}
		return result;
	}
	
	private void setElementValue(Element aElement, String aIdentifier, String aValue)
	{	if (aIdentifier.equals("text"))
		{	XMLUtils.setText(aElement, aValue);
		}
		else
		{	aElement.setAttribute(aIdentifier,aValue);
		}
	}
	
	private void processSetXPath(Element aElement, String aXPath, String aValue) throws XPathExpressionException
	{	List<Node> nodes=XPath.eval(aXPath, aElement);
		for (Node n: nodes)
		{	switch (n.getNodeType())
			{	case Node.ELEMENT_NODE:
					XMLUtils.setText((Element)n, aValue);
					break;
				case Node.ATTRIBUTE_NODE:
					n.setNodeValue(aValue);
					break;
			}
		}
	}
	
	private void processArgumentAttributes(INKFRequest aReq, Element aElement) throws NKFException
	{	NamedNodeMap nnm=aElement.getAttributes();
		for (int i=nnm.getLength()-1; i>=0; i--)
		{	Node n=nnm.item(i);
			String argName=("xrl".equals(n.getPrefix()))?n.getLocalName():n.getNodeName();
			if (argName.startsWith(ARGUMENT_ATT))
			{	String name=argName.substring(ARGUMENT_ATT.length());
				String value=n.getNodeValue();
				try
				{	aReq.addArgument(name, value);
				}
				catch (NKFException e)
				{	//argument not available so don't add it- if mandatory then grammar will catch it
					if (!isArg(value))
					{	throw e;
					}
				}
				aElement.removeAttributeNode((Attr)n);
			}
		}
	}
	
	private INKFRequest buildRequest(XMLReadable aReadable, INKFRequestContext aContext) throws Exception
	{
		Document d=XMLUtils.newDocument();
		Element docEl=d.createElement("request");
		d.appendChild(docEl);
		String identifier=aReadable.getTrimText("xrl:identifier");
		Element idEl=d.createElement("identifier");
		XMLUtils.setText(idEl, identifier);
		docEl.appendChild(idEl);
		List<Element> arguments=(List<Element>)(List)aReadable.getNodes("xrl:argument");
		for (Element argument : arguments)
		{	Element argEl=d.createElement("argument");
			NamedNodeMap nnm=argument.getAttributes();
			for (int i=0; i<nnm.getLength(); i++)
			{	Node attIn=nnm.item(i);
				argEl.setAttribute(attIn.getLocalName(),attIn.getNodeValue());
			}
			for (Node n=argument.getFirstChild(); n!=null; n=n.getNextSibling())
			{	argEl.appendChild(d.importNode(n,true));
			}
			docEl.appendChild(argEl);
		}
		
		RequestBuilder b=new RequestBuilder(docEl,aContext.getKernelContext().getKernel().getLogger());
		INKFRequest req=b.buildRequest(aContext, null, null);
		return req;
	}
		
	private static boolean isArg(String aIdentifier)
	{	return aIdentifier.startsWith("arg:");
	}
	
	private Document getMutableClone(Node aNode) throws ParserConfigurationException
	{
		Document result;
		if (aNode instanceof Document)
		{	result=(Document)XMLUtils.safeDeepClone(aNode);
		}
		else
		{	result=org.netkernel.layer0.util.XMLUtils.newDocument();
			result.appendChild(result.importNode(aNode, true));
		}
		return result;
	}
	
	private String dropScheme(String aIdentifier)
	{
		String result;
		int i=aIdentifier.indexOf(':');
		if (i>=0)
		{	result=aIdentifier.substring(i+1);
		}
		else
		{	result=aIdentifier;
		}
		return result;
	}
	
}