package org.netkernel.lang.xrl.endpoint;

import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.util.XMLReadable;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.layer0.util.XPath;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.w3c.dom.*;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPathExpressionException;
import java.util.List;


public class XRL3Runtime extends StandardAccessorImpl
  {
  // we look for control tags in either the default namespace or this namespace
  private static final String XRL_NS = "http://netkernel.org/xrl";
  private static final String ARGUMENT_ATT = "argument-";

  public XRL3Runtime()
    {
    declareThreadSafe();
    }

  public void onSource(INKFRequestContext context) throws Exception
    {
    INKFRequest request;
    Document resultXML;

    String templateId = context.getThisRequest().getArgumentValue("template");

    // Get the XML template
    request = context.createRequest(templateId);
    request.setRepresentationClass(Node.class);
    Node template = (Node) context.issueRequest(request);

    // Find out if we are to be tolerant of XRL errors
    boolean tolerant = context.getThisRequest().argumentExists("tolerant");

    resultXML = interpretXRL(template, context, tolerant);

    INKFResponse response = context.createResponseFrom(resultXML);
    String mimetype = context.getThisRequest().getArgumentValue("mimetype");
    if (mimetype != null)
      {
      response.setMimeType(mimetype);
      }
    else
      {
      response.setMimeType("application/xml");
      }
    }


  //========== Private Methods ==========

  private Document interpretXRL(Node node, INKFRequestContext context, boolean tolerant) throws Exception
    {
    Document xmlResult = getMutableClone(node);
    processXRLInElement(xmlResult.getDocumentElement(), context, tolerant);
    return xmlResult;
    }


  private void processXRLInElement(Element element, INKFRequestContext context, boolean tolerant) throws Exception
    {
    // Remove the XRL namespace from the element
    if (element.getAttribute("xmlns:xrl").length() > 0)
      {
      element.removeAttribute("xmlns:xrl");
      }

    // Process the children of the provided element
    Element e = XMLUtils.getFirstChildElement(element);
    while (e != null)
      {
      Element nextElement = XMLUtils.getNextSiblingElement(e);
      String ns = e.getNamespaceURI();
      if (ns != null && ns.equals(XRL_NS))
        {
        // If the child is an XRL tag, then we process the whole tag
        String name = e.getLocalName();
        if (name.equals("include") || name.equals("include-children"))
          {
          processInclude(e, context, tolerant);
          }
        else if (name.equals("resolve"))
          {
          processResolve(e, context, tolerant);
          }
        }
      else
        {
        // This node is not a command; examine for embedded XRL
        processEmbeddedXRL(e, context, tolerant);

        // Process the node for child XRL tags
        processXRLInElement(e, context, tolerant);
        }
      e = nextElement;
      }
    }

  private Element processEmbeddedXRL(Element element, INKFRequestContext context, boolean tolerant) throws Exception
    {
    String attribute;
    attribute = element.getAttribute("xrl:include");
    if (attribute.length() > 0)
      {
      processInlineInclude(element, attribute, context, tolerant);
      }

    attribute = element.getAttribute("xrl:resolve");
    if (attribute.length() > 0)
      {
      processInlineResolve(element, attribute, context, tolerant);
      }
    return element;
    }


  // Interpret the passed element as an XRL Include directive
  private void processInclude(Element element, INKFRequestContext context, boolean tolerant) throws Exception
    {
    String includeContentIdentifier = null;
    INKFRequest request = null;
    boolean hasXPath = false;
    String xPath = null;

    Element elementToReplace = element;  // By default we replace the XRL instruction tag
    try
      {
      if (XMLUtils.getFirstChildElement(element) == null)
        {
        // Single-element specification of the XRL Include instruction
        if (element.hasAttribute("xrl:identifier"))
          {
          includeContentIdentifier = element.getAttribute("xrl:identifier");
          request = context.createRequest(includeContentIdentifier);
          }
        else
          {
          exceptionHandler(context, tolerant, "EX_INCLUDE", "MSG_NO_BASE_OR_TARGET", element, null, includeContentIdentifier);
          return;
          }
        request = processArgumentAttributes(request, element);
        request = processRepresentationTypeAttribute(request, element);
        Object inclusion = context.issueRequest(request);

        if (inclusion instanceof String)
          {
          hasXPath = element.hasAttribute("xrl:xpath");
          xPath = element.getAttribute("xrl:xpath");
          String replacementText = (String) inclusion;

          if (!hasXPath)
            {
            xPath = ".."; // Special case for String
            }
          if (xPath.length() > 0)
            {
            processSetXPath(element, xPath, replacementText);
            // Get rid of XRL tag
            element.getParentNode().removeChild(element);
            }
          }
        else if (inclusion instanceof Node)
          {
          Node n = (Node) inclusion;
          if (n instanceof Document)
            {
            n = ((Document) n).getDocumentElement();
            Element frag = (Element) element.getOwnerDocument().importNode(n, true);
            elementToReplace.getParentNode().replaceChild(frag, elementToReplace);
            if (elementToReplace != element)
              {
              element.getParentNode().removeChild(element);
              }
            processXRLInElement(frag, context, tolerant);
            }
          }
        else
          {
          // ??
          System.out.println("Bad class");
          }
        }
      else
        {
        // Multi-element specification of the XRL Include instruction
        XMLReadable xmlReadable = new XMLReadable(element);
        includeContentIdentifier = xmlReadable.getTrimText("xrl:identifier");
        if (includeContentIdentifier.equals(""))
          {
          exceptionHandler(context, tolerant, "EX_INCLUDE", "MSG_NO_BASE_OR_TARGET", element, null, includeContentIdentifier);
          return;
          }
        request = context.createRequest(includeContentIdentifier);
        request = processRepresentationType(request, xmlReadable);
        request = processArguments(request, xmlReadable);
        Object inclusion = context.issueRequest(request);

        // TODO: What if it's a null document?

        hasXPath = element.hasAttribute("xrl:xpath");
        xPath = element.getAttribute("xrl:xpath");


        if (inclusion instanceof String)
          {
          if (!hasXPath)
            {
            xPath = ".."; // Special case for String
            }
          if (xPath.length() > 0)
            {
            String replacementText = (String) inclusion;
            processSetXPath(element, xPath, replacementText);
            // Get rid of XRL tag
            element.getParentNode().removeChild(element);
            }
          }
        else if (inclusion instanceof Node)
          {
          String xpath = element.getAttribute("xrl:xpath");
          if (xpath.length() > 0)
            {
            List<Node> nodes = XPath.eval(xpath, element);
            if (nodes.size() == 1 && nodes.get(0) instanceof Element)
              {
              elementToReplace = (Element) nodes.get(0);
              }
            else
              {
              exceptionHandler(context, tolerant, "EX_INCLUDE", "MSG_BAD_XPATH", element, null, includeContentIdentifier);
              }
            }

          Node n = (Node) inclusion;
          if (n instanceof Document)
            {
            n = ((Document) n).getDocumentElement();
            }
          Element frag = (Element) element.getOwnerDocument().importNode(n, true);
          elementToReplace.getParentNode().replaceChild(frag, elementToReplace);
          if (elementToReplace != element)
            {
            element.getParentNode().removeChild(element);
            }
          processXRLInElement(frag, context, tolerant);
          }
        else
          {
          // ??
          System.out.println("Bad class");
          }
        }
      }
    catch (Exception e)
      {
      exceptionHandler(context, tolerant, "EX_INCLUDE", "MSG_EVAL", element, e, includeContentIdentifier);
      }
    }


  // Interpret the passed element as an XRL Resolve directive
  private void processResolve(Element element, INKFRequestContext context, boolean tolerant) throws Exception
    {
    INKFRequest request;
    String identifier = null;
    String xPath;
    try
      {
      XMLReadable xml = new XMLReadable(element);
      identifier = xml.getTrimText("xrl:identifier");
      if (identifier.equals(""))
        {
        exceptionHandler(context, tolerant, "EX_RESOLVE", "MSG_NO_BASE_OR_TARGET", element, null, identifier);
        return;
        }
      request = context.createRequest(identifier);
      processArguments(request, xml);
      identifier = request.getIdentifier();
      identifier = dropScheme(identifier);
      xPath = element.getAttribute("xrl:xpath");
      processSetXPath(element, xPath, identifier);
      element.getParentNode().removeChild(element);
      }
    catch (Exception e)
      {
      exceptionHandler(context, tolerant, "EX_RESOLVE", "MSG_EVAL", element, e, identifier);
      }

    }

  private void processInlineInclude(Element element, String aEval, INKFRequestContext context, boolean tolerant) throws Exception
    {
    INKFRequest request;
    String target = "??";
    try
      {
      target = getElementValue(element, aEval, context);
      element.removeAttribute("xrl:include");
//      INKFRequest req = createRequestToEndpoint(target, context);
      request = context.createRequest(target);
      processArgumentAttributes(request, element);
      request.setRepresentationClass(String.class);
      String evalResult = (String) context.issueRequest(request);
      setElementValue(element, aEval, evalResult);
      }
    catch (Exception e)
      {
      exceptionHandler(context, tolerant, "EX_EVAL", "MSG_EVAL", element, e, target);
      }
    }

  private void processInlineResolve(Element element, String aResolve, INKFRequestContext context, boolean tolerant) throws Exception
    {
    String target = "??";

    try
      {
      target = getElementValue(element, aResolve, context);
      if (target.indexOf("[[") >= 0)
        {
        String oldtarget = null;
        int p1, p2 = 0;
        while ((p1 = target.indexOf("[[")) >= 0)
          {
          p2 = target.indexOf("]]");
          oldtarget = target;
          target = target.substring(p1 + 2, p2);
          INKFRequest req = createRequestToEndpoint(target, context);
          processArgumentAttributes(req, element);
          String identifier = req.getIdentifier();
          identifier = dropScheme(identifier);
          target = oldtarget.substring(0, p1) + identifier + oldtarget.substring(p2 + 2);
          }
        element.removeAttribute("xrl:resolve");
        }
      else
        {
        element.removeAttribute("xrl:resolve");
        INKFRequest req = createRequestToEndpoint(target, context);
        processArgumentAttributes(req, element);
        String identifier = req.getIdentifier();
        target = dropScheme(identifier);
        }
      setElementValue(element, aResolve, target);
      }
    catch (Exception e)
      {
      exceptionHandler(context, tolerant, "EX_RESOLVE", "MSG_EVAL", element, e, target);
      }
    }


  private INKFRequest processRepresentationTypeAttribute(INKFRequest request, Element element) throws NKFException, ParserConfigurationException, ClassNotFoundException
    {
    String representationClassName = element.getAttribute("xrl:representation");
    if (representationClassName.equals(""))
      {
      // We use the default
      request.setRepresentationClass(Node.class);
      }
    else
      {
      Class representationClass = Class.forName(representationClassName);
      request.setRepresentationClass(representationClass);
      }
    return request;
    }

  private INKFRequest processRepresentationType(INKFRequest request, XMLReadable element) throws NKFException, ParserConfigurationException, ClassNotFoundException
    {
    String representationClassName = element.getTrimText("xrl:representation");
    if (representationClassName.equals(""))
      {
      // We use the default
      request.setRepresentationClass(Node.class);
      }
    else
      {
      Class representationClass = Class.forName(representationClassName);
      request.setRepresentationClass(representationClass);
      }
    return request;
    }

  private INKFRequest processArgumentAttributes(INKFRequest request, Element element) throws NKFException
    {
    NamedNodeMap nnm = element.getAttributes();
    for (int i = nnm.getLength() - 1; i >= 0; i--)
      {
      Node n = nnm.item(i);
      String argName = ("xrl".equals(n.getPrefix())) ? n.getLocalName() : n.getNodeName();
      if (argName.startsWith(ARGUMENT_ATT))
        {
        String name = argName.substring(ARGUMENT_ATT.length());
        String value = n.getNodeValue();
        try
          {
          request.addArgument(name, value);
          }
        catch (NKFException e)
          {  //argument not available so don't add it- if mandatory then grammar will catch it
          if (!isArg(value))
            {
            throw e;
            }
          }
        element.removeAttributeNode((Attr) n);
        }
      }
    return request;
    }


  private INKFRequest processArguments(INKFRequest request, XMLReadable element) throws NKFException, ParserConfigurationException
    {
    // Routine augments the request by iterating through the child arguments and adding them to the request
    List<Element> arguments = (List<Element>) (List) element.getNodes("xrl:argument");
    for (Element argument : arguments)
      {
      String name = argument.getAttribute("name");
      Node literalRoot = XMLUtils.getFirstChildElement(argument);
      if (literalRoot == null)
        {
        String value = XMLUtils.getText(argument);
        try
          {
          request.addArgument(name, value);
          }
        catch (NKFException e)
          {  //argument not available so don't add it- if mandatory then grammar will catch it
          if (!isArg(value))
            {
            throw e;
            }

          }
        }
      else
        {
        Document d = XMLUtils.newDocument();
        d.appendChild(d.importNode(literalRoot, true));
        request.addArgumentByValue(name, d);
        }
      }
    return request;
    }

  private void processSetXPath(Element aElement, String aXPath, String aValue) throws XPathExpressionException
    {
    List<Node> nodes = XPath.eval(aXPath, aElement);
    for (Node n : nodes)
      {
      switch (n.getNodeType())
        {
        case Node.ELEMENT_NODE:
          XMLUtils.setText((Element) n, aValue);
          break;
        case Node.ATTRIBUTE_NODE:
          n.setNodeValue(aValue);
          break;
        }
      }
    }

  private String getElementValue(Element element, String aIdentifier, INKFRequestContext context) throws Exception
    {
    String result;
    if (aIdentifier.equals("text"))
      {
      result = XMLUtils.getText(element);
      }
    else
      {
      Attr attr = element.getAttributeNode(aIdentifier);
      if (attr != null)
        {
        result = attr.getNodeValue();
        }
      else
        {
        throw context.createFormattedException("EX_ATTRIBUTE_NOT_FOUND", "MSG_ATTRIBUTE_NOT_FOUND", null, null, aIdentifier);
        }

      }
    return result;
    }

  private void setElementValue(Element element, String identifier, String value)
    {
    if (identifier.equals("text"))
      {
      XMLUtils.setText(element, value);
      }
    else
      {
      element.setAttribute(identifier, value);
      }
    }

  private INKFRequest createRequestToEndpoint(String target, INKFRequestContext context) throws NKFException
    {
    return context.createRequestToEndpoint(target);
    }


  private Document getMutableClone(Node aNode) throws ParserConfigurationException
    {
    Document result;
    if (aNode instanceof Document)
      {
      result = (Document) XMLUtils.safeDeepClone(aNode);
      }
    else
      {
      result = org.netkernel.layer0.util.XMLUtils.newDocument();
      result.appendChild(result.importNode(aNode, true));
      }
    return result;
    }

  private void exceptionHandler(INKFRequestContext aHelper, boolean tolerant, String aException, String aMessage, Node aElement, Exception e, String target) throws Exception
    {
    if (tolerant)
      {
      aHelper.logFormatted(aHelper.LEVEL_WARNING, aMessage, XMLUtils.getPathFor(aElement), e, target);
      //e.printStackTrace();
      System.err.println(e.toString());
      }
    else
      {
      throw aHelper.createFormattedException(aException, aMessage, XMLUtils.getPathFor(aElement), e, target);
      }
    }

  private static boolean isArg(String identifier)
    {
    return identifier.startsWith("arg:");
    }

  private String dropScheme(String aIdentifier)
    {
    String result;
    int i = aIdentifier.indexOf(':');
    if (i >= 0)
      {
      result = aIdentifier.substring(i + 1);
      }
    else
      {
      result = aIdentifier;
      }
    return result;
    }

  }

